@@ -1,6 +1,6 @@
 ;;; ebdb-gnus.el --- Gnus interface to EBDB          -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2016-2018  Free Software Foundation, Inc.
+;; Copyright (C) 2016-2020  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
 
@@ -31,31 +31,40 @@
 (autoload 'message-make-domain "message")
 
 (defgroup ebdb-mua-gnus nil
-  "Gnus-specific EBDB customizations"
+  "Gnus-specific EBDB customizations."
   :group 'ebdb-mua)
 (put 'ebdb-mua-gnus 'custom-loads '(ebdb-gnus))
 
-(defcustom ebdb-gnus-window-configuration
-  `(article
-    ,(cond
-      (gnus-use-trees
-       '(vertical 1.0
-		  (summary 0.25 point)
-		  (tree 0.25)
-		  (horizontal 1.0
-			      (article 1.0)
-			      (ebdb-gnus 0.4))))
-      (t
-       '(vertical 1.0
-		  (summary 0.25 point)
-		  (horizontal 1.0
-			      (article 1.0)
-			      (ebdb-gnus 0.4))))))
-  "Gnus window configuration to include EBDB.
-By default, this adds the *EBDB-Gnus* window to the right of the
-article buffer, taking up 40% of the horizontal space."
-  :group 'ebdb-mua-gnus
-  :type 'list)
+(defcustom ebdb-gnus-auto-update-p ebdb-mua-reader-update-p
+  "Gnus-specific value of `ebdb-mua-auto-update-p'."
+  :type '(choice (const :tag "do nothing" nil)
+                 (const :tag "search for existing records" existing)
+                 (const :tag "update existing records" update)
+                 (const :tag "query for update or record creation" query)
+                 (const :tag "update or create automatically" create)
+                 (function :tag "User-defined function")))
+
+(defcustom ebdb-gnus-window-size ebdb-default-window-size
+  "Size of the EBDB buffer when popping up in Gnus.
+Size should be specified as a float between 0 and 1.  Defaults to
+the value of `ebdb-default-window-size'."
+  :type 'float)
+
+(defcustom ebdb-gnus-window-configuration nil
+  "Symbol that names EBDB's Gnus window config.
+This option is nil by default, meaning Gnus will pop up the
+*EBDB-Gnus* window next to the *Article* buffer, with a
+width/height of `ebdb-gnus-window-size'.
+
+If more control is required, set this to a symbol name.  This
+symbol will be entered into the `gnus-window-to-buffer' alist,
+and can be used as an entry in more complex Gnus buffer/window
+configurations.
+
+Note that this should be a different symbol from that used in
+Message-mode article composition window config."
+  :type '(choice (const :tag nil)
+		 (symbol :tag "Window config name")))
 
 (defcustom ebdb-gnus-post-style-function
   (lambda (_rec _mail) nil)
@@ -68,7 +77,6 @@ nil to use Gnus defaults.
 
 See `ebdb-record-field' or `ebdb-record-current-fields' for
 likely ways to extract information about the record."
-  :group 'ebdb-mua-gnus
   :type 'function)
 
 (defgroup ebdb-mua-gnus-scoring nil
@@ -95,7 +103,7 @@ likely ways to extract information about the record."
 (cl-defmethod ebdb-read ((field (subclass ebdb-gnus-score-field)) &optional slots obj)
   (let ((score (string-to-number
 		(ebdb-read-string
-		 "Score: " (when obj (slot-value obj 'score))))))
+		 "Score" (when obj (slot-value obj 'score))))))
     (cl-call-next-method field (plist-put slots :score score) obj)))
 
 (cl-defmethod ebdb-string ((field ebdb-gnus-score-field))
@@ -199,7 +207,7 @@ addresses better than the traditionally static global scorefile."
 	 (list :folder (plist-get slots :group))))
 
 (defun ebdb/gnus-split-folders-list ()
-  "Return a list of \( \"From\" mail-regexp imap-folder-name\) tuples
+  "Return a list of (\"From\" mail-regexp imap-folder-name) tuples
 based on the contents of the EBDB.
 
 Mail address elements are already `regexp-quote'-ed, so we just
@@ -210,10 +218,10 @@ use a backquote template, that is your setting will look like:
 \(setq nnimap-split-rule  'nnimap-split-fancy
        nnimap-split-inbox \"INBOX\"
        nnimap-split-fancy
-       `\(| ,@\(ebdb/gnus-split-folders-list\)
-            ... \)\)
+       `(| ,@(ebdb/gnus-split-folders-list)
+            ... ))
 
-Note that `\( is the backquote, NOT the quote '\(."
+Note that `( is the backquote, NOT the quote '(."
   (mapcar
    (lambda (elt)
      (list "From"
@@ -225,18 +233,16 @@ Note that `\( is the backquote, NOT the quote '\(."
 ;; Insinuation
 ;;
 
-(add-hook 'gnus-article-prepare-hook #'ebdb-mua-auto-update)
+(defun ebdb-gnus-auto-update ()
+  (ebdb-mua-auto-update ebdb-gnus-auto-update-p))
+
+(add-hook 'gnus-article-prepare-hook #'ebdb-gnus-auto-update)
 
 (add-hook 'gnus-startup-hook #'ebdb-insinuate-gnus)
 
 (defsubst ebdb-gnus-buffer-name ()
   (format "*%s-Gnus*" ebdb-buffer-name))
 
-;; Tell Gnus how to display the *EBDB-Gnus* buffer.
-(with-eval-after-load "gnus-win"
-  (add-to-list 'gnus-window-to-buffer
-	       `(ebdb-gnus . ,(ebdb-gnus-buffer-name)))
-  (gnus-add-configuration ebdb-gnus-window-configuration))
 
 (cl-defmethod ebdb-make-buffer-name (&context (major-mode gnus-summary-mode))
   "Produce a EBDB buffer name associated with Gnus."
@@ -252,10 +258,10 @@ Note that `\( is the backquote, NOT the quote '\(."
 	   (unless (gnus-buffer-live-p gnus-article-buffer)
 	     (gnus-summary-show-article))
 	   (get-buffer-window gnus-article-buffer))))
-    (list win 0.3)))
+    (list win ebdb-gnus-window-size)))
 
 (cl-defmethod ebdb-popup-window (&context (major-mode gnus-article-mode))
-  (list (get-buffer-window) 0.3))
+  (list (get-buffer-window) ebdb-gnus-window-size))
 
 ;; It seems that `gnus-fetch-field' fetches decoded content of
 ;; `gnus-visible-headers', ignoring `gnus-ignored-headers'.
@@ -338,6 +344,20 @@ composed to a certain record."
   (define-key gnus-summary-mode-map ";" ebdb-mua-keymap)
   (define-key gnus-article-mode-map ";" ebdb-mua-keymap)
 
+  ;; If the user has set this to a symbol, it needs to be added to
+  ;; Gnus' `gnus-window-to-buffer' list.
+  (when ebdb-gnus-window-configuration
+    (add-to-list 'gnus-window-to-buffer
+		 (cons ebdb-gnus-window-configuration
+		       (ebdb-gnus-buffer-name))))
+
+  ;; Versions of Gnus with the gnus-search.el library allow us to
+  ;; perform contact auto-completion within search queries.
+  (when (boundp 'gnus-search-contact-tables)
+    (add-hook 'ebdb-after-load-hook
+	      (lambda ()
+		(push ebdb-hashtable gnus-search-contact-tables))))
+
   ;; Set up user field for use in `gnus-summary-line-format'
   ;; (1) Big solution: use whole name
   (if ebdb-mua-summary-unify-format-letter
