@@ -1,6 +1,6 @@
 ;;; ebdb-migrate.el --- Migration/upgrade functions for EBDB  -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2016-2018  Free Software Foundation, Inc.
+;; Copyright (C) 2016-2020  Free Software Foundation, Inc.
 
 ;; Author: Eric Abrahamsen <eric@ericabrahamsen.net>
 
@@ -25,12 +25,15 @@
 ;;; Code:
 
 (require 'ebdb)
+(require 'cl-lib)
+(require 'subr-x)
 (autoload 'calendar-absolute-from-gregorian "calendar")
 (autoload 'calendar-gregorian-from-absolute "calendar")
 (autoload 'org-time-string-to-absolute "org")
 (declare-function make-gnorb-ebdb-link "ext:gnorb-ebdb")
 (declare-function org-contacts-db "ext:org-contacts")
 (defvar url-handler-regexp)
+(defvar bbdb-time-stamp-format)
 ;;; Migrating the EBDB
 
 ;; For org-contacts
@@ -59,55 +62,24 @@ Format ((VERSION . DIFFERENCES) ... ).")
 ;; Quiet compiler
 (defvar bbdb-file)
 
-;;; These struct functions and definitions are here only to enable
-;;; multi-step upgrades from earlier versions of the database.
-
-(defmacro ebdb-defstruct (name &rest elts)
-  "Define two functions to operate on vector NAME for each symbol ELT in ELTS.
-The function ebdb-NAME-ELT returns the element ELT in vector NAME.
-The function ebdb-NAME-set-ELT sets ELT.
-Also define a constant ebdb-NAME-length that holds the number of ELTS
-in vector NAME."
-  (declare (indent 1))
-  (let* ((count 0)
-         (sname (symbol-name name))
-         (uname (upcase sname))
-         (cname (concat "ebdb-" sname "-"))
-         body)
-    (dolist (elt elts)
-      (let* ((selt (symbol-name elt))
-             (setname  (intern (concat cname "set-" selt))))
-        (push (list 'defsubst (intern (concat cname selt)) `(,name)
-                    (format "For EBDB %s read element %i `%s'."
-                            uname count selt)
-                    ;; Use `elt' instead of `aref' so that these functions
-                    ;; also work for the `ebdb-record-type' pseudo-code.
-                    `(elt ,name ,count)) body)
-        (push (list 'defsubst setname `(,name value)
-                    (format "For EBDB %s set element %i `%s' to VALUE.  \
-Return VALUE.
-Do not call this function directly.  Call instead `ebdb-record-set-field'
-which ensures the integrity of the database.  Also, this makes your code
-more robust with respect to possible future changes of EBDB's innermost
-internals."
-                            uname count selt)
-                    `(aset ,name ,count value)) body))
-      (setq count (1+ count)))
-    (push (list 'defconst (intern (concat cname "length")) count
-                (concat "Length of EBDB `" sname "'.")) body)
-    (cons 'progn body)))
-
 ;; Define structs so that the getters/setters used elsewhere in the
 ;; file operate normally.  These functions are used nowhere else in
 ;; EBDB, and the "v" prefix has been added to prevent function name
 ;; clashes.
-(ebdb-defstruct vrecord
-  firstname lastname affix aka organization phone address mail xfields cache)
+(cl-defstruct (ebdb-vrecord
+               (:type vector)
+               (:constructor nil))
+ firstname lastname affix aka organization phone address mail xfields
+ uuid creation-date timestamp cache)
 
-(ebdb-defstruct vphone
+(cl-defstruct (ebdb-vphone
+               (:type vector)
+               (:constructor nil))
   label area exchange suffix extension)
 
-(ebdb-defstruct vaddress
+(cl-defstruct (ebdb-vaddress
+               (:type vector)
+               (:constructor nil))
   label streets city state postcode country)
 
 (defun ebdb-peel-the-onion (lis)
@@ -121,7 +93,7 @@ slightly munged old EBDB files."
     (setq lis (car lis)))
   lis)
 
-(defconst ebdb-file-format 7
+(defconst ebdb-file-format 9
   "Obsolete variable, only used in migration.")
 
 ;;;###autoload
@@ -142,22 +114,38 @@ slightly munged old EBDB files."
                                (elt record 5) (elt record 6) (elt record 7)
                                (elt record 8)))
           (setq temp (cdr temp)))))
+  (if (< old-format 9)
+      (let (new-records)
+        (dolist (record records)
+          (let ((uuid (or (cdr (assq 'uuid (elt record 8)))
+                          (ebdb-make-uuid)))
+                (creation-date (or (cdr (assq 'creation-date (elt record 8)))
+                                   (format-time-string bbdb-time-stamp-format nil t)))
+                (timestamp (or (cdr (assq 'timestamp (elt record 8)))
+                               (format-time-string bbdb-time-stamp-format nil t))))
+            (push (vector (elt record 0) (elt record 1) (elt record 2)
+                          (elt record 3) (elt record 4) (elt record 5)
+                          (elt record 6) (elt record 7)
+                          (let ((xfields (elt record 8)))
+                            (dolist (elt '(uuid creation-date timestamp))
+                              (setq xfields (assq-delete-all elt xfields)))
+                            xfields)
+                          uuid creation-date timestamp
+                          (elt record 9))
+                  new-records)))
+        (setq records (nreverse new-records))))
   (mapc (ebdb-migrate-versions-lambda old-format) records)
   records)
 
 (defconst ebdb-migration-spec
-  '((2 (ebdb-vrecord-xfields ebdb-vrecord-set-xfields
-        ebdb-migrate-change-dates))
-    (3 (ebdb-vrecord-address ebdb-vrecord-set-address
-        ebdb-migrate-add-country-field))
-    (4 (ebdb-vrecord-address ebdb-vrecord-set-address
-        ebdb-migrate-streets-to-list))
-    (5 (ebdb-vrecord-address ebdb-vrecord-set-address
-        ebdb-migrate-postcodes-to-strings))
-    (6 (ebdb-vrecord-xfields ebdb-vrecord-set-xfields
-        ebdb-migrate-xfields-to-list)
-       (ebdb-vrecord-organization ebdb-vrecord-set-organization
-        ebdb-migrate-organization-to-list)))
+  '((2 (ebdb-vrecord-xfields ebdb-migrate-change-dates))
+    (3 (ebdb-vrecord-address ebdb-migrate-add-country-field))
+    (4 (ebdb-vrecord-address ebdb-migrate-streets-to-list))
+    (5 (ebdb-vrecord-address ebdb-migrate-postcodes-to-strings))
+    (6 (ebdb-vrecord-xfields ebdb-migrate-xfields-to-list)
+       (ebdb-vrecord-organization ebdb-migrate-organization-to-list))
+    (7 (bbdb-record-xfields bbdb-migrate-xfields-to-list)
+       (bbdb-record-organization bbdb-migrate-organization-to-list)))
   "The alist of (version . migration-spec-list).
 See `ebdb-migrate-record-lambda' for details.")
 
@@ -172,12 +160,11 @@ modified, and SET is the function to be used to set the field to be
 modified.  FUNCTION will be applied to the result of GET, and its
 results will be saved with SET."
   (byte-compile `(lambda (record)
-                  ,@(mapcar (lambda (ch)
-                              `(,(cadr ch) record
-                                (,(car (cddr ch))
-                                 (,(car ch) record))))
-                            changes)
-                  record)))
+                   ,@(mapcar (lambda (ch)
+                               `(cl-callf ,(cadr ch)
+                                    (,(car ch) record)))
+                             changes)
+                   record)))
 
 (defun ebdb-migrate-versions-lambda (v0)
   "Return the function to migrate from V0 to `ebdb-file-format'."
@@ -236,7 +223,7 @@ This uses the code that used to be in `ebdb-address-postcode'."
                                   ;; else a number, could be 3 to 5 digits (possible error: assuming
                                   ;; no leading zeroes in postcodes)
                                   (format "%d" (ebdb-vaddress-postcode address)))))))
-              (ebdb-vaddress-set-postcode address postcode))
+              (setf (ebdb-vaddress-postcode address) postcode))
             address)
           addresses))
 
@@ -339,13 +326,13 @@ Formats are changed in timestamp and creation-date fields from
       (ebdb-split 'organization organization)
     organization))
 
-;;; These defcustoms are now obsolete, but they're here so that,
-;;; during the migration/upgrade process, we know which xfields to
-;;; handle specially, and turn into specific field types.  In the case
-;;; of `bbdb/gnus-split-public-field', this should signal to us that
-;;; the record should actually be changed into a
-;;; `ebdb-record-mailing-list'.  But that hasn't been implemented yet,
-;;; so...
+;; These defcustoms are now obsolete, but they're here so that,
+;; during the migration/upgrade process, we know which xfields to
+;; handle specially, and turn into specific field types.  In the case
+;; of `bbdb/gnus-split-public-field', this should signal to us that
+;; the record should actually be changed into a
+;; `ebdb-record-mailing-list'.  But that hasn't been implemented yet,
+;; so...
 
 (defcustom bbdb/gnus-split-private-field 'gnus-private
   "This variable is used to determine the xfield to reference to find the
@@ -383,7 +370,6 @@ checked for a score to add to the mail addresses in the same record."
 ;;;###autoload
 (defun ebdb-migrate-from-bbdb (&optional file)
   "Migrate from BBDB to EBDB.
-
 This upgrade is extreme enough that we can't really use the
 existing migration mechanisms.  They are still there, though, in
 case someone's going from, say, version 2 to 4 in one jump.
@@ -391,7 +377,7 @@ case someone's going from, say, version 2 to 4 in one jump.
 Migrate from FILE, if non-nil.  Otherwise, assume that the
 variable `bbdb-file' points to an existing file holding valid
 contacts in a previous BBDB format.  If that variable isn't set
-use \(locate-user-emacs-file \"bbdb\" \".bbdb\"\), which is how
+use (locate-user-emacs-file \"bbdb\" \".bbdb\"), which is how
 BBDB sets the default of that option."
   (interactive)
   (require 'url-handlers)
@@ -403,7 +389,7 @@ BBDB sets the default of that option."
 				 bbdb-file)
 			    (locate-user-emacs-file "bbdb" ".bbdb")))
     (when (and (/= (point-min) (point-max))
-	       (yes-or-no-p "Upgrade from previous version of BBDB? "))
+	       (yes-or-no-p "Migrate records from BBDB? "))
       (let ((v-records (ebdb-migrate-parse-records))
 	    (target-db (ebdb-prompt-for-db nil t))
 	    (total 0)
@@ -414,15 +400,16 @@ BBDB sets the default of that option."
 	      (let ((orgs (ebdb-vrecord-organization r))
 		    (c-rec (ebdb-migrate-vector-to-class r))
 		    org)
-		;; Gives it a uuid
+		;; Gives it a uuid if it doesn't already have one.
 		(ebdb-db-add-record target-db c-rec)
 		(when orgs
 		  (dolist (o orgs)
 		    ;; Make all the orgs into real records.
-		    (unless (string= o "") ; There are many of these.
+		    (unless (string-blank-p o) ; There are many of these.
 		      (setq org (or (seq-find
 				     (lambda (r)
-				       (string= o (ebdb-record-name r)))
+				       (and (ebdb-record-organization-p r)
+					    (string= o (ebdb-record-name-string r))))
 				     c-records)
 				    (let ((time (current-time)))
 				      (ebdb-db-add-record
@@ -476,17 +463,31 @@ BBDB sets the default of that option."
 	(address (aref v 6))
 	(mail (aref v 7))
 	(xfields (aref v 8))
-	name akas phones mails addresses fields ts c-date notes lab val)
-    (setq name (make-instance ebdb-default-name-class
-			      :surname l-name
-			      :given-names (when f-name (split-string f-name " " nil))
-			      :affix affix))
+	(uuid (aref v 9))
+	(c-date (aref v 10))
+	(ts (aref v 11))
+	name akas phones mails addresses fields notes lab val
+	instance)
+    (setq name (if (child-of-class-p ebdb-default-record-class
+				     'ebdb-record-person)
+		   (make-instance
+		    ebdb-default-name-class
+		    :surname l-name
+		    :given-names (when f-name (split-string f-name " " nil))
+		    :affix affix)
+		 (make-instance 'ebdb-field-name-simple
+				:name (if f-name
+					  (concat f-name " " l-name)
+					l-name))))
     (when aka
       (dolist (a aka)
 	(push (make-instance ebdb-default-name-class
 			     :surname (car (last (split-string a)))
 			     :given-names (butlast (split-string a)))
 	      akas)))
+    (when uuid
+      (setq uuid
+	    (make-instance 'ebdb-field-uuid :uuid uuid)))
     (when phone
       (dolist (p phone)
 	(let ((label (aref p 0))
@@ -532,13 +533,18 @@ BBDB sets the default of that option."
 		  mails))))
       (when mails
 	(setf (slot-value (car (last mails)) 'priority) 'primary)))
-    (when xfields
-      (dolist (x xfields)
-	(setq lab (car x)
-	      val (cdr x))
-	(cond
-	 ((eq lab 'timestamp)
-	  (setq ts (make-instance 'ebdb-field-timestamp
+    (when ts
+      (setq ts (make-instance 'ebdb-field-timestamp
+			      :timestamp
+			      (apply #'encode-time
+				     (mapcar
+				      (lambda (el)
+					(if (null el)
+					    0
+					  el))
+				      (parse-time-string ts))))))
+    (when c-date
+      (setq c-date (make-instance 'ebdb-field-creation-date
 				  :timestamp
 				  (apply #'encode-time
 					 (mapcar
@@ -546,23 +552,19 @@ BBDB sets the default of that option."
 					    (if (null el)
 						0
 					      el))
-					  (parse-time-string val))))))
-	 ((eq lab 'creation-date)
-	  (setq c-date (make-instance 'ebdb-field-creation-date
-				      :timestamp
-				      (apply #'encode-time
-					     (mapcar
-					      (lambda (el)
-						(if (null el)
-						    0
-						  el))
-					      (parse-time-string val))))))
+					  (parse-time-string c-date))))))
+    (when xfields
+      (dolist (x xfields)
+	(setq lab (car x)
+	      val (cdr x))
+	(cond
 	 ((eq lab 'mail-alias)
-	  (push (make-instance 'ebdb-field-mail-alias
-			       :alias val
-			       :address (car-safe mails))
-		fields))
-	 ((and (stringp v)
+	  (when mails
+	    (push (make-instance 'ebdb-field-mail-alias
+				 :alias val
+				 :address (car-safe mails))
+		  fields)))
+	 ((and (stringp val)
 	       (string-match-p val url-handler-regexp))
 	  (push (make-instance 'ebdb-field-url
 			       :url val)
@@ -615,17 +617,23 @@ BBDB sets the default of that option."
 			       :label (symbol-name (car x))
 			       :value val)
 		fields)))))
-    (make-instance ebdb-default-record-class
-		   :name name
-		   :aka akas
-		   :phone phones
-		   :address addresses
-		   :mail mails
-		   :timestamp ts
-		   :creation-date c-date
-		   :notes notes
-		   :dirty t
-		   :fields fields)))
+    (setq instance
+	  (make-instance ebdb-default-record-class
+			 :name name
+			 :uuid uuid
+			 :phone phones
+			 :address addresses
+			 :mail mails
+			 :timestamp ts
+			 :creation-date c-date
+			 :notes notes
+			 :dirty t
+			 :fields fields))
+    (when akas
+      (condition-case nil
+	  (setf (slot-value instance 'aka) akas)
+	(invalid-slot-type nil)))
+    instance))
 
 (defun ebdb-migrate-parse-records ()
   "Parse an earlier (non-EIEIO) version of a BBDB database file."
